from wata.pointcloud.utils import utils
from wata.pointcloud.utils import move_pcd
from wata.pointcloud.utils import o3d_visualize_utils
from numpy import ndarray
from pathlib import Path
from typing import Union

class PointCloudProcess:

    @staticmethod
    def cut_pcd(points:ndarray, pcd_range:list, axis:str):
        '''
        **功能描述**: 切割指定范围的点云  
        
        Args:  
            points: numpy格式的点云  
            pcd_range: [x1,y1,z1,x2,y2,z2]  
            axis: 可选 "xy","xz","yz","xz","xyz",默认为"xyz"  

        Returns:  
            返回范围内的点  
        '''
        return utils.cut_pcd(points, pcd_range, axis)

    @staticmethod
    def filter_points(points:ndarray, del_points:ndarray):
        '''
        **功能描述**: 过滤删除部分点,要求del_points是points的子集  
        
        Args:  
            points: numpy格式的点云  
            del_points: 要从points中删除的点,numpy格式的点云  

        Returns:  
            剩余的点,numpy格式  
        '''
        return utils.filter_points(points, del_points)

    @staticmethod
    def show_pcd(path:Union[str, Path], point_size:float=1, background_color:list=None, pcd_range:list=None, bin_num_features:int=None,
                 create_coordinate:bool=True, create_plane:bool=False, type:str='open3d'):
        '''
        **功能描述**: 直接可视化点云文件,目前支持bin pcd npy 格式的点云  
        
        Args:
            path: 点云的路径  
            point_size: 点的大小  
            background_color: 背景颜色  
            pcd_range: 可视化点云的范围  
            bin_num_features: 如果是bin格式的文件,应给出bin格式点云的列数  
            create_coordinate: 是否可视化坐标系  
            create_plane: 是否可视化出一个平面  
            type: 用什么工具可视化, 默认open3d 可选 matplotlib open3d qtopengl mayavi(未开发) vispy(未开发)  

        Returns:  
            直接弹出可视化界面  
        '''
        utils.show_pcd(path, point_size, background_color, pcd_range, bin_num_features,
                       create_coordinate, create_plane, type)

    @staticmethod
    def points_to_o3d_model(points:ndarray, point_colors:Union[None, list]=None):
        '''
        **功能描述**: 将numpy格式的点云转化为open3d模型  
        
        Args:
            points: numpy格式的点云  
            point_colors: 点云的颜色信息,可以为None,可以为长度为3的元素为int类型的RGB列表  

        Returns:  
            open3d模型  
        '''
        return o3d_visualize_utils.points_to_o3d_model(points=points, point_colors=point_colors)

    @staticmethod
    def save_o3d_camera_parameters(points:ndarray, save_file_path:Union[str, Path]='camera_parameters.json', window_size:list=[1200, 800]):
        '''
        **功能描述**:  保存手动调整的open3d的视角  
        
        Args:  
            points: numpy格式的点云  
            save_file_path: open3d 可视化视角文件保存的路径 json  
            window_size: 可视化窗口的大小 list  

        Returns:  
            无  
        '''
        return o3d_visualize_utils.save_o3d_camera_parameters(points, save_file_path, window_size)

    @staticmethod
    def show_pcd_from_points(points:Union[ndarray,list], point_size:float=1, background_color:list=None, colors:Union[list, int]=None, create_coordinate=True,
                             create_plane=False, type='open3d'):
        '''
        **功能描述**: 展示点云  
        
        Args:  
            points: numpy格式的点云;当points为list时代表将list中的numpy格式的点云全部可视化  
            point_size: 可视化点的大小,默认为1  
            background_color: 背景颜色  
            colors: 点云的颜色,当colors为int类型时,为用点云的第几列显色;当colors为list类型时,colors代表[R,G,B]  
            create_coordinate: 是否可视化坐标系  
            create_plane: 是否可视化出一个平面  
            type: 用什么工具可视化, 默认open3d 可选 matplotlib open3d qtopengl mayavi(未开发) vispy(未开发)  

        Returns:  
            无  
        '''
        utils.show_pcd_from_points(points, point_size, background_color, colors, create_coordinate, create_plane, type)

    @staticmethod
    def get_points(path:str, num_features:Union[None, int]=None):
        '''
        **功能描述**: 读取点云文件获取点
        
        Args:
            path: 点云文件的路径支持 pcd npy bin 格式的点云  
            num_features: 当点云文件格式为pcd 或npy时,num_features为加载点的列数; 当点云文件格式为bin时,num_features为bin文件点云的列数  

        Returns:  
            返回numpy格式的点云矩阵  
        '''
        return utils.get_points(path, num_features)

    @staticmethod
    def add_boxes(points:Union[ndarray,list], gt_boxes:Union[ndarray,None]=None, gt_labels:Union[None, list]=None, pred_boxes:Union[ndarray,None]=None, pred_labels:Union[None, list]=None, pred_scores:Union[None, list]=None,
                  point_size:int=1,
                  background_color:list=None, create_plane:bool=False, point_colors:Union[None, list]=None, create_coordinate=True, type='open3d',
                  savepath=None, plot_range=None):
        '''
        **功能描述**: 绘制点云、标注框、预测框  
        
        Args:  
            points: numpy格式的点云;当points为list时代表将list中的numpy格式的点云全部可视化  
            gt_boxes: GT框  
            gt_labels: GT框标签  
            pred_boxes: 预测框  
            pred_labels: 预测框标签  
            pred_scores: 预测框的分数  
            point_size: 可视化点的大小,默认为1  
            background_color: 背景颜色  
            create_plane: 是否创建一个可视化地面  
            point_colors: 对应points为列表时的颜色,当points为numpy格式时不起作用  
            create_coordinate: 是否在原点创建坐标系  
            type: 用什么工具可视化, 默认open3d 可选 matplotlib open3d qtopengl mayavi(未开发) vispy(未开发)  
            savepath: 当type为matplotlib时,保存图像的地址  
            plot_range: 当type为matplotlib时,可视化的范围  

        Returns:  
            可视化界面  
        '''
        utils.add_boxes(points, gt_boxes=gt_boxes, gt_labels=gt_labels, pred_boxes=pred_boxes, pred_labels=pred_labels,
                        pred_scores=pred_scores, point_size=point_size,
                        background_color=background_color, create_plane=create_plane, point_colors=point_colors,
                        create_coordinate=create_coordinate, type=type, savepath=savepath, plot_range=plot_range)

    @staticmethod
    def pcd2bin(pcd_dir:Union[str, Path], bin_dir:Union[str, Path], num_features=4):
        '''
        **功能描述**: pcd格式点云转bin格式  
        
        Args:  
            pcd_dir: pcd格式点云数据的存放目录  
            bin_dir: bin格式点云的数据村帆帆目录  
            num_features: bin格式的点云列数  

        Returns:  
            无  
        '''
        utils.pcd2bin(pcd_dir, bin_dir, num_features)

    @staticmethod
    def xyzrpy2RTmatrix(xyz_rpy, degrees=False):
        '''
        **功能描述**: xyzrpy2RTmatrix  
        
        Args:
            xyz_rpy: list [dx, dy, dz, roll, pitch, yaw]  
            degrees: roll, pitch, yaw 是弧度还是角度 默认为弧度(False)  

        Returns:  
            numpy格式的4x4旋转平移矩阵   
        '''
        return move_pcd.xyzrpy2RTmatrix(xyz_rpy=xyz_rpy, degrees=degrees)

    @staticmethod
    def RTmatrix2xyzrpy(RTmatrix:ndarray):
        '''
        **功能描述**: RTmatrix2xyzrpy  
        
        Args:  
            RTmatrix: numpy格式的4x4旋转平移矩阵  

        Returns:  
            numpy格式1x6矩阵 分别为 dx, dy, dz, roll, pitch, yaw  
        '''
        return move_pcd.RTmatrix2xyzrpy(RTmatrix)

    @staticmethod
    def move_pcd_with_RTmatrix(points:ndarray, RTmatrix:ndarray, inv=False):
        '''
        **功能描述**: 通过旋转平移矩阵移动点云  
        
        Args:
            points: numpy格式的点云  
            RTmatrix: numpy格式的4x4旋转平移矩阵  
            inv: 是否对 RTmatrix 取逆,默认为False  

        Returns:  
            旋转平移后的numpy格式的点云  
        '''
        return move_pcd.move_pcd_with_RTmatrix(points, RTmatrix, inv)

    @staticmethod
    def move_pcd_with_xyzrpy(points:ndarray, xyz_rpy, degrees=False):
        '''
        **功能描述**: 用 dx, dy, dz, roll, pitch, yaw 旋转平移点云  
        
        Args:  
            points: numpy格式的点云,需要有x,y,z维度  
            xyz_rpy: 列表格式[dx, dy, dz, roll, pitch, yaw]  
            degrees: xyz_rpy 中的roll, pitch, yaw是否为角度值,默认为False  

        Returns:  
            返回旋转平移后的点云  
        '''
        return move_pcd.move_pcd_with_xyzrpy(points, xyz_rpy, degrees=degrees)

    @staticmethod
    def cartesian_to_spherical(points:ndarray):
        '''
        **功能描述**: 直角坐标系转笛卡尔坐标系  
        
        Args:  
            points: 直角坐标系点云  

        Returns:
            (r, theta, phi)的numpy格式点云  
        '''
        return utils.cartesian_to_spherical(points=points)

    @staticmethod
    def get_v_channel_from_pcd(points:ndarray, vfov, channel_nums, offset=0.01):
        '''
        **功能描述**: 获取垂直方向的通道  
        
        Args:  
            points: numpy格式的点云,需要有x,y,z维度  
            vfov: 长度为2的列表,代表了垂直视场角的范围,水平方向为0度,低于水平方向为负  
            channel_nums: 垂直通道的线数  
            offset: 偏移量  

        Returns:  
            numpy格式的 nx1 的数组  
        '''
        return utils.get_v_channel_from_pcd(points, vfov, channel_nums, offset)

    @staticmethod
    def get_h_channel_from_pcd(points:ndarray, hfov, channel_nums, offset=0.001):
        '''
        **功能描述**: 获取点云水平方向的通道  
        
        Args:
            points: numpy格式的点云,需要有x,y,z维度  
            hfov: 长度为2的列表,代表了水平视场角的范围,前向120度的视场角可用[30,150]表示  
            channel_nums: 水平通道的线数  
            offset: 偏移量  

        Returns:
            numpy格式的 nx1 的数组  
        '''
        return utils.get_h_channel_from_pcd(points, hfov, channel_nums, offset)

    @staticmethod
    def points_in_boxes(points:ndarray, boxes, type="gpu"):
        '''
        **功能描述**: 此API需要安装pytorch,用于标记和查找点云中在指定框内的点  
        
        Args:
            points: numpy格式的点云,需要有x,y,z维度  
            boxes: nx7的numpy矩阵  
            type: 可选"gpu" "cpu"  

        Returns:
            返回每个点所在框的索引列表  
        '''
        return utils.points_in_boxes(points, boxes, type)

    @staticmethod
    def save_pcd(points:ndarray, save_path, fields=None, npdtype=None, type='binary'):
        '''
        **功能描述**: 保存pcd格式的点云  
        
        Args:
            points: numpy格式的点云  
            save_path: pcd格式的点云文件保存的路径  
            fields: 点云每一列的元信息,当为None时默认为["x","y","z","4","5",....]  
            npdtype: 点云每一列保存的数据类型,支持numpy格式和字符串格式。字符串格式可选 "f32","f64","i8","i16","i32","i64","u8","u16","u32","u64"  
            type: 有两种格式,分别为“binary” 和 “ascii”,默认为“binary”。binary格式保存为二进制文件,ascii保存为普通文件  

        Returns:  
            无  
        '''
        return utils.save_pcd(points, save_path, fields, npdtype, type)



from wata.display.utils import utils

class Display:
    
    @staticmethod
    def print(content,type="r"):
        '''
        **功能描述**: 特色打印
        
        Args:
            content: 内容  
            type: 类型支持 字母r g b y p 代表颜色,rr表示粗体,r_表示红色下划线,rx表示红色斜体  

        Returns:
            无  
        '''
        return utils.wataprint(content,type)


import torch
import torch.nn as nn
from torch.autograd import Function
import numpy as np

from . import roiaware_pool3d_cuda

def check_numpy_to_torch(x):
    if isinstance(x, np.ndarray):
        return torch.from_numpy(x).float(), True
    return x, False

def points_in_boxes_cpu(points, boxes):
    """
    Args:
        points: (num_points, 3)
        boxes: [x, y, z, dx, dy, dz, heading], (x, y, z) is the box center, each box DO NOT overlaps
    Returns:
        point_indices: (N, num_points)
    """
    assert boxes.shape[1] == 7
    assert points.shape[1] == 3
    points, is_numpy = check_numpy_to_torch(points)
    boxes, is_numpy = check_numpy_to_torch(boxes)

    point_indices = points.new_zeros((boxes.shape[0], points.shape[0]), dtype=torch.int)
    roiaware_pool3d_cuda.points_in_boxes_cpu(boxes.float().contiguous(), points.float().contiguous(), point_indices)

    return point_indices.numpy() if is_numpy else point_indices


def points_in_boxes_gpu(points, boxes):
    """
    :param points: (B, M, 3)
    :param boxes: (B, T, 7), num_valid_boxes <= T
    :return box_idxs_of_pts: (B, M), default background = -1
    """
    assert boxes.shape[0] == points.shape[0]
    assert boxes.shape[2] == 7 and points.shape[2] == 3
    batch_size, num_points, _ = points.shape

    box_idxs_of_pts = points.new_zeros((batch_size, num_points), dtype=torch.int).fill_(-1)
    roiaware_pool3d_cuda.points_in_boxes_gpu(boxes.contiguous(), points.contiguous(), box_idxs_of_pts)

    return box_idxs_of_pts


class RoIAwarePool3d(nn.Module):
    def __init__(self, out_size, max_pts_each_voxel=128):
        super().__init__()
        self.out_size = out_size
        self.max_pts_each_voxel = max_pts_each_voxel

    def forward(self, rois, pts, pts_feature, pool_method='max'):
        assert pool_method in ['max', 'avg']
        return RoIAwarePool3dFunction.apply(rois, pts, pts_feature, self.out_size, self.max_pts_each_voxel, pool_method)


class RoIAwarePool3dFunction(Function):
    @staticmethod
    def forward(ctx, rois, pts, pts_feature, out_size, max_pts_each_voxel, pool_method):
        """
        Args:
            ctx:
            rois: (N, 7) [x, y, z, dx, dy, dz, heading] (x, y, z) is the box center
            pts: (npoints, 3)
            pts_feature: (npoints, C)
            out_size: int or tuple, like 7 or (7, 7, 7)
            max_pts_each_voxel:
            pool_method: 'max' or 'avg'

        Returns:
            pooled_features: (N, out_x, out_y, out_z, C)
        """
        assert rois.shape[1] == 7 and pts.shape[1] == 3
        if isinstance(out_size, int):
            out_x = out_y = out_z = out_size
        else:
            assert len(out_size) == 3
            for k in range(3):
                assert isinstance(out_size[k], int)
            out_x, out_y, out_z = out_size

        num_rois = rois.shape[0]
        num_channels = pts_feature.shape[-1]
        num_pts = pts.shape[0]

        pooled_features = pts_feature.new_zeros((num_rois, out_x, out_y, out_z, num_channels))
        argmax = pts_feature.new_zeros((num_rois, out_x, out_y, out_z, num_channels), dtype=torch.int)
        pts_idx_of_voxels = pts_feature.new_zeros((num_rois, out_x, out_y, out_z, max_pts_each_voxel), dtype=torch.int)

        pool_method_map = {'max': 0, 'avg': 1}
        pool_method = pool_method_map[pool_method]
        roiaware_pool3d_cuda.forward(rois, pts, pts_feature, argmax, pts_idx_of_voxels, pooled_features, pool_method)

        ctx.roiaware_pool3d_for_backward = (pts_idx_of_voxels, argmax, pool_method, num_pts, num_channels)
        return pooled_features

    @staticmethod
    def backward(ctx, grad_out):
        """
        :param grad_out: (N, out_x, out_y, out_z, C)
        :return:
            grad_in: (npoints, C)
        """
        pts_idx_of_voxels, argmax, pool_method, num_pts, num_channels = ctx.roiaware_pool3d_for_backward

        grad_in = grad_out.new_zeros((num_pts, num_channels))
        roiaware_pool3d_cuda.backward(pts_idx_of_voxels, argmax, grad_out.contiguous(), grad_in, pool_method)

        return None, None, grad_in, None, None, None


if __name__ == '__main__':
    pass



import math
import utm
import re

X_PI = 3.14159265358979324 * 3000.0 / 180.0
PI = 3.1415926535897932384626
A = 6378245.0
EE = 0.00669342162296594323


def transformLat(x, y):
    ret = -100.0 + 2.0 * x + 3.0 * y + 0.2 * y * y + 0.1 * x * y + 0.2 * math.sqrt(abs(x))
    ret += (20.0 * math.sin(6.0 * x * PI) + 20.0 * math.sin(2.0 * x * PI)) * 2.0 / 3.0
    ret += (20.0 * math.sin(y * PI) + 40.0 * math.sin(y / 3.0 * PI)) * 2.0 / 3.0
    ret += (160.0 * math.sin(y / 12.0 * PI) + 320.0 * math.sin(y * PI / 30.0)) * 2.0 / 3.0
    return ret


def transformLon(x, y):
    ret = 300.0 + x + 2.0 * y + 0.1 * x * x + 0.1 * x * y + 0.1 * math.sqrt(abs(x))
    ret += (20.0 * math.sin(6.0 * x * PI) + 20.0 * math.sin(2.0 * x * PI)) * 2.0 / 3.0
    ret += (20.0 * math.sin(x * PI) + 40.0 * math.sin(x / 3.0 * PI)) * 2.0 / 3.0
    ret += (150.0 * math.sin(x / 12.0 * PI) + 300.0 * math.sin(x / 30.0 * PI)) * 2.0 / 3.0
    return ret


def WGS84_to_GCJ02(lat, lon):
    if out_of_china(lat, lon):
        return lat, lon
    dLat = transformLat(lon - 105.0, lat - 35.0)
    dLon = transformLon(lon - 105.0, lat - 35.0)
    radLat = lat / 180.0 * PI
    magic = math.sin(radLat)
    magic = 1 - EE * magic * magic
    sqrtMagic = math.sqrt(magic)
    dLat = (dLat * 180.0) / ((A * (1 - EE)) / (magic * sqrtMagic) * PI)
    dLon = (dLon * 180.0) / (A / sqrtMagic * math.cos(radLat) * PI)
    mgLat = lat + dLat
    mgLon = lon + dLon
    return mgLat, mgLon


def GCJ02_to_WGS84(lat, lon):
    GCJ02 = WGS84_to_GCJ02(lat, lon)
    dLat = GCJ02[0] - lat
    dLon = GCJ02[1] - lon
    return lat - dLat, lon - dLon


def GCJ02_to_BD09(lat, lon):
    z = math.sqrt(lon * lon + lat * lat) + 0.00002 * math.sin(lat * X_PI)
    theta = math.atan2(lat, lon) + 0.000003 * math.cos(lon * X_PI)
    bd_lon = z * math.cos(theta) + 0.0065
    bd_lat = z * math.sin(theta) + 0.006
    return bd_lat, bd_lon


def BD09_to_GCJ02(lat, lon):
    x = lon - 0.0065
    y = lat - 0.006
    z = math.sqrt(x * x + y * y) - 0.00002 * math.sin(y * X_PI)
    theta = math.atan2(y, x) - 0.000003 * math.cos(x * X_PI)
    gcj_lon = z * math.cos(theta)
    gcj_lat = z * math.sin(theta)
    return gcj_lat, gcj_lon


def WGS84_to_BD09(lat, lon):
    GCJ02 = WGS84_to_GCJ02(lat, lon)
    return GCJ02_to_BD09(GCJ02[0], GCJ02[1])


def BD09_to_WGS84(lat, lon):
    GCJ02 = BD09_to_GCJ02(lat, lon)
    return GCJ02_to_WGS84(GCJ02[0], GCJ02[1])


# def cgcs2000_to_WGS84(lat, lon):
#     return lat, lon  # Assume CGCS2000 is equivalent to WGS84 for simplicity
#
#
# def WGS84_to_cgcs2000(lat, lon):
#     return lat, lon  # Assume CGCS2000 is equivalent to WGS84 for simplicity


# def GCJ02_to_cgcs2000(lat, lon):
#     WGS84 = GCJ02_to_WGS84(lat, lon)
#     return WGS84_to_cgcs2000(WGS84[0], WGS84[1])
#
#
# def cgcs2000_to_GCJ02(lat, lon):
#     WGS84 = cgcs2000_to_WGS84(lat, lon)
#     return WGS84_to_GCJ02(WGS84[0], WGS84[1])


# WGS84 to UTM
def WGS84_to_UTM(lat, lon):
    u = utm.from_latlon(lat, lon)
    return u


# UTM to WGS84
def UTM_to_WGS84(easting, northing, zone_number, zone_letter):
    latlon = utm.to_latlon(easting, northing, zone_number, zone_letter)
    return latlon


def dms_to_decimal(dms_lat, dms_lon):
    lat = convert_dms_to_decimal(dms_lat)
    lon = convert_dms_to_decimal(dms_lon)
    return lat, lon


def convert_dms_to_decimal(dms):
    degrees, minutes, seconds, direction = re.split('[°\'"]', dms)
    decimal = float(degrees) + float(minutes) / 60 + float(seconds) / 3600
    if direction in ['S', 'W']:
        decimal *= -1
    return decimal


def decimal_to_dms(decimal_lat, decimal_lon):
    lat_dms = convert_decimal_to_dms(decimal_lat, True)
    lon_dms = convert_decimal_to_dms(decimal_lon, False)
    return lat_dms, lon_dms


def convert_decimal_to_dms(decimal, is_latitude):
    direction = 'N' if is_latitude else 'E'
    if decimal < 0:
        direction = 'S' if is_latitude else 'W'
    decimal = abs(decimal)
    degrees = int(decimal)
    minutes = int((decimal - degrees) * 60)
    seconds = (decimal - degrees - minutes / 60) * 3600
    return f'{degrees}°{minutes}\'{seconds:.2f}"{direction}'


def out_of_china(lat, lon):
    return lon < 72.004 or lon > 137.8347 or lat < 0.8293 or lat > 55.8271


import yaml
import json
import pickle
import os


def load_yaml(path):
    with open(path, 'r', encoding='UTF-8') as f:
        yaml_datas = yaml.load(f.read(), Loader=yaml.FullLoader)
    return yaml_datas


def write_yaml(data, save_path):
    yaml_data = yaml.dump(data)
    with open(save_path, 'w', encoding='UTF-8') as f:
        f.write(yaml_data)


def load_json(path):
    with open(path, 'r', encoding='UTF-8') as f:
        json_datas = json.loads(f.read())
    return json_datas

def load_geojson(path):
    with open(path, 'r', encoding='UTF-8') as f:
        json_datas = json.loads(f.read())
    return json_datas


def write_json(data, save_path):
    json_data = json.dumps(data, indent=4, ensure_ascii=False)
    with open(save_path, 'w', encoding='UTF-8') as f:
        f.write(json_data)


def load_pkl(path):
    with open(path, 'rb') as f:
        pkl_datas = pickle.load(f)
    return pkl_datas


def write_pkl(data, save_path):
    with open(save_path, 'wb') as f:
        pickle.dump(data, f)


def load_txt(path):
    with open(path, 'r', encoding='UTF-8') as f:
        lines = f.readlines()
        content = [line.rstrip() for line in lines]
    return content


def write_txt(data, save_path):
    with open(save_path, 'w', encoding='UTF-8') as f:
        if isinstance(data,str):
            f.write(data)
        elif isinstance(data,list):
            for res in data:
                if str(res)[-1:] != '\n':
                    f.write(str(res) + '\n')
                else:
                    f.write(str(res))
        else:
            raise ValueError("Only str and list can be selected for data")


import os
from pathlib import Path


def recursive_file_path(path):
    path_dict = {}
    num_dict = {}
    for foldername, subfolders, filenames in os.walk(path):
        if not subfolders:
            num_dict[foldername] = len(filenames)

        display_filenames = [os.path.join(foldername, file) for file in filenames[:2]]
        if len(filenames) > 10:
            display_filenames.append(os.path.join(foldername, '......'))

        display_subfolders = [os.path.join(foldername, folder) for folder in subfolders[:15]]
        path_dict[foldername] = [display_subfolders, display_filenames]

    return path_dict, num_dict


def recursion(dict, key, output, num, root_name):
    for k, i in enumerate(dict[key][0]):  # dir
        datai = os.path.relpath(i, root_name)
        datai_sp = datai.split(os.sep)
        if i in num.keys() and num[i] > 80:
            output.append((len(datai_sp) - 1) * '│  ' + "├── " + "{:20}".format(datai_sp[-1]) + "({})".format(num[i]))
        else:
            output.append((len(datai_sp) - 1) * '│  ' + "├── " + datai_sp[-1])

        recursion(dict, i, output, num, root_name)
    for k, i in enumerate(dict[key][1]):  # file
        datai = os.path.relpath(i, root_name)
        datai_sp = datai.split(os.sep)
        if k == len(dict[key][1]) - 1:
            output.append((len(datai_sp) - 1) * '│  ' + "└── " + datai_sp[-1])
        else:
            output.append((len(datai_sp) - 1) * '│  ' + "├── " + datai_sp[-1])


def file_tree(datasets_path, save_path=None):
    root_name = datasets_path
    dict, num = recursive_file_path(datasets_path)
    output = [Path(root_name).name]
    recursion(dict, datasets_path, output, num, root_name)
    for i in output:
        print(i)

    if save_path is not None:
        with open(save_path, 'w', encoding='utf-8') as f:
            for line in output:
                f.write(line + '\n')
        print('\n', save_path + " has been saved!")


if __name__ == '__main__':
    file_tree("D:\Code\wata")



import os
from pathlib import Path
import numpy as np
from wata.file.utils import load_and_write_file


def load_file(path):
    file_ext = Path(path).suffix[1:]
    if file_ext in ['yaml', 'json', 'geojson', 'pkl', 'txt']:
        return eval('load_and_write_file.load_' + file_ext)(path)
    else:
        raise NameError("Unable to handle {} formatted files".format(file_ext))

def save_file(data, save_path):
    file_ext = Path(save_path).suffix[1:]
    if file_ext in ['yaml', 'json', 'pkl', 'txt']:
        return eval('load_and_write_file.write_' + file_ext)(data, save_path)
    else:
        raise NameError("Unable to handle {} formatted files".format(type))

def mkdir_if_not_exist(path):
    if not os.path.exists(path):
        os.makedirs(path)

def np2str(arr, return_type='list'):
    if isinstance(arr, list):
        arr = np.array(arr)
    if arr.ndim == 1:
        string = ' '.join(map(str, arr))
        return string

    elif arr.ndim == 2:
        string_list = []
        for arr_line in arr:
            string = ' '.join(map(str, arr_line)) + "\n"
            string_list.append(string)
        if return_type =='str':
            string = str(string_list)[2:-2].replace("', '", "").replace("\\n", "\n")
            return string
        elif return_type =='list':
            return string_list
        else:
            raise ValueError("Only 'str' and 'list' can be selected for return_type")
    else:
        raise ValueError("Only supports arrays of 2 dimensions and below")


"""
Open3d visualization tool box
Written by Jihan YANG
All rights preserved from 2021 - present.
"""
import matplotlib
import numpy as np
import open3d
import os
from matplotlib import pyplot as plt

gt_box_colormap = [
    [0, 0.5, 0],
    [0, 0.5, 0.5],
    [0.5, 0.5, 0],
    [0.5, 0, 0.5],
    [0.5, 0, 0],
    [0, 0, 0.5],
    [0.5, 0.5, 0.5],
]
per_box_colormap = [
    [0, 1, 0],
    [0, 1, 1],
    [1, 1, 0],
    [1, 0, 1],
    [1, 0, 0],
    [0, 0, 1],
    [1, 1, 1],
]


def create_mesh_plane(grid_width=10.0, plane_width=200.0, z_position=-2.2, color=[0.2, 0.2, 0.2]):
    lines = []
    line_points = []
    grid_nums = plane_width / grid_width / 2
    for j in range(int(-1 * grid_nums), int(grid_nums + 1)):
        line_points.extend([[plane_width / 2, -j * grid_width, z_position],
                            [-1 * plane_width / 2, -j * grid_width, z_position]])
        # print("-----------------")
        # for j in range(int(-1 * grid_nums), int(grid_nums + 1)):
        line_points.extend([[j * grid_width, plane_width / 2, z_position],
                            [j * grid_width, -1 * plane_width / 2, z_position]])
    for i in range(int(len(line_points) / 2)):
        lines.append([2 * i, 2 * i + 1])
    line_set = open3d.geometry.LineSet()
    line_set.points = open3d.utility.Vector3dVector(line_points)
    line_set.lines = open3d.utility.Vector2iVector(lines)
    line_set.colors = open3d.utility.Vector3dVector([color for i in range(len(lines))])  # 设置颜色为红色
    # o3d.visualization.draw_geometries([line_set])
    return line_set


def get_coor_colors(obj_labels):
    """
    Args:
        obj_labels: 1 is ground, labels > 1 indicates different instance cluster

    Returns:
        rgb: [N, 3]. color for each point.
    """
    colors = matplotlib.colors.XKCD_COLORS.values()
    max_color_num = obj_labels.max()

    color_list = list(colors)[: max_color_num + 1]
    colors_rgba = [matplotlib.colors.to_rgba_array(color) for color in color_list]
    label_rgba = np.array(colors_rgba)[obj_labels]
    label_rgba = label_rgba.squeeze()[:, :3]

    return label_rgba


def points_to_o3d_model(points, point_colors=None):
    pts = open3d.geometry.PointCloud()
    pts.points = open3d.utility.Vector3dVector(points[:, :3])
    if point_colors is not None:
        if isinstance(point_colors, list) and isinstance(point_colors[0], int):
            point_colors = np.tile(point_colors, (len(points), 1))
        pts.colors = open3d.utility.Vector3dVector(point_colors)
    return pts


def open3d_draw_scenes(points, colors=None, gt_boxes=None, gt_labels=None, pred_boxes=None, pred_labels=None,
                       pred_scores=None,
                       point_size=1, background_color=None, create_plane=True,
                       create_coordinate=True, cam_param=None, window_size=[1200, 800]):
    o3d_model_list = []
    if isinstance(points, list):
        for i, single_points in enumerate(points):
            single_points_pcd = points_to_o3d_model(points)
            if colors is not None:
                assert isinstance(colors, list), "color should be list"
                assert len(points) == len(colors), "The length of points color should be the same"
                points_colors = np.tile(colors[i], (len(single_points), 1))
                single_points_pcd.colors = open3d.utility.Vector3dVector(points_colors)
            o3d_model_list.append(single_points_pcd)
    else:
        pts = points_to_o3d_model(points)
        o3d_model_list.append(pts)

    if create_plane:
        plane = create_mesh_plane()
        o3d_model_list.append(plane)

    if create_coordinate:
        coordinate = open3d.geometry.TriangleMesh.create_coordinate_frame(size=1.0, origin=[0, 0, 0])
        o3d_model_list.append(coordinate)

    if gt_boxes is not None:
        gt_boxes_lineset_list = draw_box(gt_boxes, (0.5, 0.5, 0.5), gt_labels, gt_box_colormap)
        o3d_model_list.extend(gt_boxes_lineset_list)

    if pred_boxes is not None:
        pred_boxes_lineset_list = draw_box(pred_boxes, (0, 1, 0), pred_labels, per_box_colormap, pred_scores)
        o3d_model_list.extend(pred_boxes_lineset_list)

    show_o3d_model(o3d_model_list, point_size, background_color, cam_param=cam_param, window_size=window_size)


def translate_boxes_to_open3d_instance(gt_boxes):
    """
       4-------- 6
     /|         /|
    5 -------- 3 .
    | |        | |
    . 7 -------- 1
    |/         |/
    2 -------- 0
    """
    center = gt_boxes[0:3]
    lwh = gt_boxes[3:6]
    axis_angles = np.array([0, 0, gt_boxes[6] + 1e-10])
    rot = open3d.geometry.get_rotation_matrix_from_axis_angle(axis_angles)
    box3d = open3d.geometry.OrientedBoundingBox(center, rot, lwh)

    line_set = open3d.geometry.LineSet.create_from_oriented_bounding_box(box3d)

    # import ipdb; ipdb.set_trace(context=20)
    lines = np.asarray(line_set.lines)
    lines = np.concatenate([lines, np.array([[1, 4], [7, 6]])], axis=0)

    line_set.lines = open3d.utility.Vector2iVector(lines)

    return line_set, box3d


def draw_box(gt_boxes, color=(1, 0, 0), labels=None, box_colormap=[[1, 0, 1], [1, 0, 0]], score=None):
    box_model_list = []
    for i in range(gt_boxes.shape[0]):
        line_set, box3d = translate_boxes_to_open3d_instance(gt_boxes[i])
        if labels is None:
            line_set.paint_uniform_color(color)
            box_model_list.append(line_set)
        else:
            line_set.paint_uniform_color(box_colormap[labels[i]])
            box_model_list.append(line_set)

        # if score is not None:
        #     corners = box3d.get_box_points()
        #     vis.add_3d_label(corners[5], '%.2f' % score[i])
    return box_model_list


def show_pcd_from_points_by_open3d(points, point_size=1, background_color=None, colors=None, create_coordinate=True,
                                   create_plane=True, cam_param=None, window_size=[1200, 800]):
    o3d_model_list = []
    if isinstance(colors, int):
        if isinstance(points, list):
            points = np.vstack(points)
        # print(points.shape)
        points_pcd = open3d.geometry.PointCloud()
        points_pcd.points = open3d.utility.Vector3dVector(points[:,:3])
        colors = points[:, colors]
        # colors[colors < 50] = 1
        colors_normalized = np.minimum(1, colors / 255)
        # Colors = plt.get_cmap('RdBu')
        # colors_normalized = Colors(colors / 50)
        # colors_normalized = (colors - np.min(colors)) / (np.max(colors) - np.min(colors))
        colors_rgb = np.stack(( colors_normalized,np.zeros_like(colors_normalized), 1-colors_normalized), axis=-1)
        points_pcd.colors = open3d.utility.Vector3dVector(colors_rgb)
        o3d_model_list.append(points_pcd)

    else:
        if isinstance(points, list):
            for i, single_points in enumerate(points):
                single_points = single_points[:, :3]
                single_points_pcd = open3d.geometry.PointCloud()
                single_points_pcd.points = open3d.utility.Vector3dVector(single_points)

                if colors is not None:
                    assert isinstance(colors, list), "color should be list"
                    assert len(points) == len(colors), "The length of points color should be the same"
                    points_colors = np.tile(colors[i], (len(single_points), 1))
                    single_points_pcd.colors = open3d.utility.Vector3dVector(points_colors)
                o3d_model_list.append(single_points_pcd)
        else:
            points = points[:, :3]
            pcd = open3d.geometry.PointCloud()
            pcd.points = open3d.utility.Vector3dVector(points)
            if colors is not None:
                points_colors = np.tile(colors, (len(points), 1))
                pcd.colors = open3d.utility.Vector3dVector(points_colors)
            o3d_model_list.append(pcd)

    if create_coordinate:
        coordinate = open3d.geometry.TriangleMesh.create_coordinate_frame(size=3)
        o3d_model_list.append(coordinate)
    if create_plane:
        plane = create_mesh_plane()
        o3d_model_list.append(plane)

    show_o3d_model(o3d_model_list, point_size=point_size, background_color=background_color, cam_param=cam_param,
                   window_size=window_size)


def load_param(vis, param_path):
    """
    读取自定义点云显示视角
    """
    param = open3d.io.read_pinhole_camera_parameters(param_path)
    ctr = vis.get_view_control()
    ctr.convert_from_pinhole_camera_parameters(param)


def show_o3d_model(o3d_model_list, point_size, background_color, cam_param=None, window_size=[1200, 800]):
    vis = open3d.visualization.Visualizer()
    vis.create_window(window_name='show_pcd_by_open3d', width=window_size[0], height=window_size[1])

    opt = vis.get_render_option()
    opt.point_size = point_size

    background_color = [0, 0, 0] if background_color is None else background_color
    opt.background_color = np.asarray(background_color)

    for model in o3d_model_list:
        vis.add_geometry(model)
    if cam_param is not None:
        load_param(vis, cam_param)
    vis.run()
    vis.clear_geometries()
    vis.destroy_window()


def save_o3d_camera_parameters(points, save_file_path='camera_parameters.json', window_size=[1200, 800]):
    vis = open3d.visualization.Visualizer()
    vis.create_window(window_name='show_pcd_by_open3d', width=window_size[0], height=window_size[1])
    opt = vis.get_render_option()
    opt.point_size = 1
    opt.background_color = np.asarray([0, 0, 0])

    points = points[:, :3]
    pcd = open3d.geometry.PointCloud()
    pcd.points = open3d.utility.Vector3dVector(points)
    vis.add_geometry(pcd)
    coordinate = open3d.geometry.TriangleMesh.create_coordinate_frame(size=3)
    vis.add_geometry(coordinate)
    plane = create_mesh_plane()
    vis.add_geometry(plane)
    # print(os.path.splitext(save_file_path))
    assert os.path.splitext(save_file_path)[-1] == '.json'
    param = vis.get_view_control().convert_to_pinhole_camera_parameters()
    open3d.io.write_pinhole_camera_parameters(save_file_path, param)

    vis.run()
    vis.clear_geometries()
    vis.destroy_window()



from matplotlib import pyplot as plt
from matplotlib.patches import Polygon
import numpy as np
from pyquaternion import Quaternion
import matplotlib

_label_colors = ['k', 'b', 'g', 'c', 'm', 'y', 'r']


def draw_single_bbox(center, size, yaw, color, linewidth):
    """
        convert box size to corners polygon
                   X
                   |
                ___|___
               |\  |  /|
               | \ | / |
        Y______|__\|/  |
               |       |
               |       |
               |_______|
    """
    # start from top-left by clockwise
    dx, dy, dz = size

    tl_p = [dx / 2.0, dy / 2.0, dz / 2.0]
    tr_p = [dx / 2.0, -dy / 2.0, dz / 2.0]
    br_p = [-dx / 2.0, -dy / 2.0, dz / 2.0]
    bl_p = [-dx / 2.0, dy / 2.0, dz / 2.0]

    # TIPS:
    rotate_q = Quaternion(axis=[0, 0, 1], radians=yaw)

    rotated_p = []
    for point3d in list([tl_p, tr_p, br_p, bl_p]):
        rotated_point3d = rotate_q.rotate(point3d)
        rotated_p.append((rotated_point3d + center)[:2])

    # print('rotated p shape', rotated_p[0].shape)
    return Polygon(
        (rotated_p[0], rotated_p[1], rotated_p[2],
         rotated_p[3], rotated_p[0], center[:2], rotated_p[1]),
        edgecolor=color,
        facecolor="none",
        linewidth=linewidth,
    )


def darw_bboxes(ax, scene_bbox, labels=None, scores=None, gt=False):
    color = 'r' if gt else 'g'
    for i in range(scene_bbox.shape[0]):
        box = scene_bbox[i]
        linewidth = 1.0
        if labels is not None:
            label = labels[i]
            linewidth = 1.5
            if not gt:
                if label < 0 or label > len(_label_colors) - 1:
                    color = "g"
                else:
                    color = _label_colors[label]

            ax.text(box[0], box[1] + 1, str(label), fontsize=20, color=color, ha='center', va='center')

        if scores is not None:
            score = scores[i]
            ax.text(box[0], box[1], str(score), fontsize=30, color='black', ha='center', va='center')

        poly = draw_single_bbox(box[:3], box[3:6], box[6], color, linewidth)
        ax.add_patch(poly)


def onscroll(event, ax):
    x_center = event.xdata
    y_center = event.ydata

    if event.button == 'up':
        ax.set_xlim(x_center - (x_center - ax.get_xlim()[0]) * 1.1, x_center + (ax.get_xlim()[1] - x_center) * 1.1)
        ax.set_ylim(y_center - (y_center - ax.get_ylim()[0]) * 1.1, y_center + (ax.get_ylim()[1] - y_center) * 1.1)
    elif event.button == 'down':
        ax.set_xlim(x_center - (x_center - ax.get_xlim()[0]) * 0.9, x_center + (ax.get_xlim()[1] - x_center) * 0.9)
        ax.set_ylim(y_center - (y_center - ax.get_ylim()[0]) * 0.9, y_center + (ax.get_ylim()[1] - y_center) * 0.9)

    plt.draw()


def ondrag(event, ax, fig):
    if event.button == 3 and event.xdata is not None and event.ydata is not None:
        x_start, y_start = event.xdata, event.ydata
        xlims = ax.get_xlim()
        ylims = ax.get_ylim()

        def onmove(event):
            nonlocal x_start, y_start
            if event.inaxes == ax:
                dx = event.xdata - x_start
                dy = event.ydata - y_start
                ax.set_xlim(xlims[0] - dx, xlims[1] - dx)
                ax.set_ylim(ylims[0] - dy, ylims[1] - dy)
                plt.draw()

        def onrelease(event):
            fig.canvas.mpl_disconnect(ondrag.id)
            fig.canvas.mpl_disconnect(onrelease.id)

        ondrag.id = fig.canvas.mpl_connect('motion_notify_event', onmove)
        onrelease.id = fig.canvas.mpl_connect('button_release_event', onrelease)


def show_pcd_from_points_by_matplotlib(points, point_size=0.6, background_color=None, colors=None,
                                       create_coordinate=True, savepath=None, plot_range=None):
    figsize = (12, 8) if savepath is None else (40, 30)
    fig, ax = plt.subplots(1, 1, figsize=figsize)

    ax.plot(0, 0, "x", color="black")
    background_color = [1, 1, 1] if background_color is None else background_color
    ax.set_facecolor(background_color)

    if create_coordinate:
        ax.arrow(0, 0, 5, 0, head_width=1, head_length=1.2, fc='r', ec='r')
        ax.arrow(0, 0, 0, 5, head_width=1, head_length=1.2, fc='g', ec='g')

    if isinstance(points, list):
        all_points = np.concatenate(points, axis=0)
        points_list = []
        if colors is not None:
            for i, single_points in enumerate(points):
                single_points = np.transpose(single_points, (1, 0)).astype(np.float32)
                points_list.append(single_points[:, :3])
                ax.scatter(single_points[0, :], single_points[1, :], c=colors[i], s=point_size)
            all_points = np.transpose(all_points, (1, 0)).astype(np.float32)
        else:
            all_points = np.transpose(all_points, (1, 0)).astype(np.float32)
            dists = np.sqrt(np.sum(all_points[:2, :] ** 2, axis=0), dtype=np.float32)
            colors = np.minimum(1, dists / 50)
            ax.scatter(all_points[0, :], all_points[1, :], c=colors, s=point_size)
    else:
        all_points = np.transpose(points, (1, 0)).astype(np.float32)
        if colors is None:
            dists = np.sqrt(np.sum(all_points[:2, :] ** 2, axis=0), dtype=np.float32)
            colors = np.minimum(1, dists / 50)
            # Colors = plt.get_cmap('Greens')  # 参考:https://zhuanlan.zhihu.com/p/114420786
            # colors = Colors(dists / 50)
        ax.scatter(all_points[0, :], all_points[1, :], c=colors, s=point_size)

    # Limit visible range.
    if plot_range is None:
        x_min = np.min(all_points[0, :])
        x_max = np.max(all_points[0, :])
        y_min = np.min(all_points[1, :])
        y_max = np.max(all_points[1, :])
        ax.set_xlim(x_min, x_max)
        ax.set_ylim(y_min, y_max)
    else:
        ax.set_xlim(plot_range[0], plot_range[2])
        ax.set_ylim(plot_range[1], plot_range[3])

    ax.tick_params(axis='x', labelsize=20)
    ax.tick_params(axis='y', labelsize=20)
    ax.set_aspect('equal')

    plt.title(savepath)
    if savepath is not None:
        plt.savefig(savepath)
        plt.close()
    else:
        matplotlib.use('TkAgg')
        fig.canvas.mpl_connect('button_press_event', lambda event: ondrag(event, ax, fig))
        fig.canvas.mpl_connect('scroll_event', lambda event: onscroll(event, ax))
        plt.show()


def plot_draw_scenes(points, gt_boxes=None, gt_labels=None, pred_boxes=None, pred_labels=None, pred_scores=None,
                     point_size=1, background_color=None, point_colors=None, create_coordinate=True, savepath=None,
                     plot_range=None):
    """
        plot the gt and dt in BEV
    """
    matplotlib.use("Agg")
    pc_points = np.transpose(points, (1, 0)).astype(np.float32)
    fig, ax = plt.subplots(1, 1, figsize=(40, 30))
    # plot ego vehicle.
    ax.plot(0, 0, "x", color="black")
    background_color = [1, 1, 1] if background_color is None else background_color
    ax.set_facecolor(background_color)

    # 在0，0位置处设置一个坐标系
    if create_coordinate:
        ax.arrow(0, 0, 5, 0, head_width=1, head_length=1.2, fc='r', ec='r')
        ax.arrow(0, 0, 0, 5, head_width=1, head_length=1.2, fc='g', ec='g')

    # Limit visible range.
    if plot_range is None:
        x_min = np.min(pc_points[0, :])
        x_max = np.max(pc_points[0, :])
        y_min = np.min(pc_points[1, :])
        y_max = np.max(pc_points[1, :])
        ax.set_xlim(x_min, x_max)
        ax.set_ylim(y_min, y_max)
    else:
        ax.set_xlim(plot_range[0], plot_range[2])
        ax.set_ylim(plot_range[1], plot_range[3])
    ax.tick_params(axis='x', labelsize=20)
    ax.tick_params(axis='y', labelsize=20)
    ax.set_aspect('equal')

    # change color with distance
    dist = np.sqrt(np.sum(pc_points[:2, :] ** 2, axis=0), dtype=np.float32)
    colors = np.minimum(1, dist / 50) if point_colors is None else plt.get_cmap(point_colors)(
        dist / 50)  # 参考:https://zhuanlan.zhihu.com/p/114420786

    ax.scatter(pc_points[0, :], pc_points[1, :], c=colors, s=point_size)

    # plot gt_bboxes
    if gt_boxes is not None:
        darw_bboxes(ax, gt_boxes, gt_labels, gt=True)

    if pred_boxes is not None:
        darw_bboxes(ax, pred_boxes, pred_labels, pred_scores, gt=False)

    plt.title(savepath)
    if savepath is not None:
        plt.savefig(savepath)
        plt.close()
    else:
        fig.canvas.mpl_connect('button_press_event', lambda event: ondrag(event, ax, fig))
        fig.canvas.mpl_connect('scroll_event', lambda event: onscroll(event, ax))
        plt.show()
